# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implement the SyncedCollection class."""
from abc import abstractmethod
from collections import defaultdict
from collections.abc import Collection
from inspect import isabstract
from threading import RLock
from typing import Any, DefaultDict, List

from ..numpy_utils import _convert_numpy
from ..utils import AbstractTypeResolver, _CounterContext, _NullContext

# Identifies types of SyncedCollection, which are the base type for this class.
_sc_resolver = AbstractTypeResolver(
    {
        "SYNCEDCOLLECTION": lambda obj: isinstance(obj, SyncedCollection),
    }
)

_collection_resolver = AbstractTypeResolver(
    {
        "COLLECTION": lambda obj: isinstance(obj, Collection),
    }
)


class _LoadAndSave:
    """A context manager for :class:`SyncedCollection` to wrap saving and loading.

    Any write operation on a synced collection must be preceded by a load and
    followed by a save. Moreover, additional logic may be required to handle
    other aspects of the synchronization, particularly the acquisition of thread
    locks. This class abstracts this concept, making it easy for subclasses to
    customize the behavior if needed (for instance, to introduce additional locks).
    """

    def __init__(self, collection):
        self._collection = collection

    def __enter__(self):
        self._collection._thread_lock.__enter__()
        self._collection._load()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._collection._save()
        finally:
            self._collection._thread_lock.__exit__(exc_type, exc_val, exc_tb)


class SyncedCollection(Collection):
    """An abstract :class:`~collections.abc.Collection` type that is synced with a backend.

    This class extends :py:class:`collections.abc.Collection` and adds a number of abstract
    internal methods that must be implemented by its subclasses. These methods can be
    split into two groups of functions that are designed to be implemented by
    separate subtrees in the inheritance hierarchy that can then be composed:

        **Concrete Collection Types**

        These subclasses should implement the APIs for specific types of
        collections. For instance, a list-like :class:`SyncedCollection`
        should implement the standard methods for sequences. In addition, they
        must implement the following abstract methods defined by the
        :class:`SyncedCollection`:

        - :meth:`~.is_base_type`: Determines whether an object satisfies the
          semantics of the collection object a given :class:`SyncedCollection`
          is designed to mimic.
        - :meth:`~._to_base`: Converts a :class:`SyncedCollection` to its
          natural base type (e.g. a `list`).
        - :meth:`~._update`: Updates the :class:`SyncedCollection` to match the
          contents of the provided :py:class:`collections.abc.Collection`.
          After calling ``sc._update(c)``, we must have that ``sc == c``; however,
          since such updates are frequent when loading and saving data to a
          resource, :meth:`_update` should be implemented to minimize new object
          creation wherever possible.

        **Backend**

        These subclasses encode the process by which in-memory data is
        converted into a representation suitable for a particular backend. For
        instance, a JSON backend should know how to save a Python object into a
        JSON-encoded file and then read that object back.

        - :meth:`~._load_from_resource`: Loads data from the underlying
          resource and returns it in an object satisfying :meth:`~.is_base_type`.
        - :meth:`~._save_to_resource`: Stores data to the underlying resource.
        - :attr:`~._backend`: A unique string identifier for the resource backend.

    Since these functionalities are effectively completely orthogonal, members of
    a given group should be interchangeable. For instance, a dict-like SyncedCollection
    can be combined equally easily with JSON, MongoDB, or SQL backends.

    **Validation**

    Due to the restrictions of a particular backend or the needs of a particular
    application, synced collections may need to restrict the data that they can
    store. Validators provide a standardized mechanism for this. A validator is
    a callable that parses any data added to a :class:`SyncedCollection` and
    raises an `Exception` if any invalid data is provided. Validators cannot
    modify the data and should have no side effects. They are purely provided
    as a mechanism to reject invalid data.  For example, a JSON validator would
    raise Exceptions if it detected non-string keys in a dict.

    Since :class:`SyncedCollection` is designed for extensive usage of
    inheritance, validators may be inherited by subclasses. There are two attributes
    that subclasses of :class:`SyncedCollection` can define to control the
    validators used:

        - ``_validators``: A list of callables that will be inherited by all
          subclasses.
        - ``_all_validators``: A list of callables that will be used to
          validate this class, and this class alone.

    When a  :class:`SyncedCollection` subclass is initialized (note that this
    is at *class* definition time, not when instances are created), its
    :meth:`_register_validators` method will be called. If this class defines
    an ``_all_validators`` attribute, this set of validators will be used by all
    instances of this class. Otherwise, :meth:`_register_validators` will traverse
    the MRO and collect the ``_validators`` attributes from all parents of a class,
    and store these in the ``_all_validators`` attribute for the class.


    .. note::

        Typically, a synced collection will be initialized with resource information,
        and data will be pulled from that resource. However, initializing with
        both data and resource information is a valid use case. In this case, the
        initial data will be validated by the standard validators, however, it
        will not be checked against the contents stored in the synced resource and
        is assumed to be consistent. This constructor pattern can be useful to
        avoid unnecessary resource accesses.

    **Thread safety**

    Whether or not :class:`SyncedCollection` objects are thread-safe depends on the
    implementation of the backend. Thread-safety of SyncedCollection objects
    is predicated on backends providing an atomic write operation. All concrete
    collection types use mutexes to guard against concurrent write operations,
    while allowing read operations to happen freely. The validity of this mode
    of access depends on the write operations of a SyncedCollection being
    atomic, specifically the `:meth:`~._save_to_resource` method. Whether or not
    a particular subclass of :class:`SyncedCollection` is thread-safe should be
    indicated by that subclass setting the ``_supports_threading`` class variable
    to ``True``. This variable is set to ``False`` by :class:`SyncedCollection`,
    so subclasses must explicitly opt-in to support threading by setting this
    variable to ``True``.

    Backends that support multithreaded execution will have multithreaded
    support turned on by default. This support can be enabled or disabled using
    the :meth:`enable_multithreading` and :meth:`disable_multithreading`
    methods. :meth:`enable_multithreading` will raise a `ValueError` if called
    on a class that does not support multithreading.


    Parameters
    ----------
    parent : SyncedCollection, optional
        If provided, the collection within which this collection is nested
        (Default value = None).
        A parent instance of :class:`SyncedCollection` or ``None``. If ``None``,
        the collection owns its own data, otherwise it is nested within its
        parent. Every :class:`SyncedCollection` either owns its own data, or has
        a parent (Default value = None).

    """

    registry: DefaultDict[str, List[Any]] = defaultdict(list)
    # Backends that support threading should modify this flag.
    _supports_threading: bool = False
    _LoadSaveType = _LoadAndSave

    def __init__(self, parent=None, *args, **kwargs):
        # Nested collections need to know their root collection, which is
        # responsible for all synchronization, and therefore all the associated
        # context managers are also stored from the root.

        if parent is not None:
            root = parent._root if parent._root is not None else parent
            self._root = root
            self._suspend_sync = root._suspend_sync
            self._load_and_save = root._load_and_save
        else:
            self._root = None
            self._suspend_sync = _CounterContext()
            self._load_and_save = self._LoadSaveType(self)

        if self._supports_threading:
            with self._cls_lock:
                if self._lock_id not in self._locks:
                    self._locks[self._lock_id] = RLock()

    @classmethod
    def _register_validators(cls):
        """Register all inherited validators to this class.

        This method is called by __init_subclass__ when subclasses are created
        to control what validators will be applied to data added to instances of
        that class. By default, the ``_all_validators`` class variable defined
        on the class itself determines the validation rules for that class. If
        that variable is not defined, then all parents of the class are searched,
        and a list of validators is constructed by concatenating the ``_validators``
        class variable for each parent class that defines it.
        """
        # Must explicitly look in cls.__dict__ so that the attribute is not
        # inherited from a parent class.
        if "_all_validators" not in cls.__dict__:
            validators = []
            # Classes inherit the validators of their parent classes.
            for base_cls in cls.__mro__:
                if hasattr(base_cls, "_validators"):
                    validators.extend(
                        [v for v in base_cls._validators if v not in validators]
                    )
            cls._all_validators = validators

    @classmethod
    def __init_subclass__(cls):
        """Register and enable validation in subclasses.

        All subclasses are given a ``_validators`` list so that separate sets of
        validators can be registered to different types of synced collections. Concrete
        subclasses (those that have all methods implemented, i.e. that are associated
        with both a specific backend and a concrete data type) are also recorded in
        an internal registry that is used to convert data from some collection-like
        object into a :class:`SyncedCollection`.
        """
        # The Python data model promises that __init_subclass__ will be called
        # after the class namespace is fully defined, so at this point we know
        # whether we have a concrete subclass or not.
        if not isabstract(cls):
            SyncedCollection.registry[cls._backend].append(cls)

        cls._register_validators()

        # Monkey-patch subclasses that support locking.
        if cls._supports_threading:
            cls._cls_lock = RLock()
            cls._locks = {}
            cls.enable_multithreading()
        else:
            cls.disable_multithreading()

    @classmethod
    def enable_multithreading(cls):
        """Enable safety checks and thread locks required for thread safety.

        Support for multithreaded execution can be disabled by calling
        :meth:`~.disable_multithreading`; calling this method reverses that.

        """
        if cls._supports_threading:

            @property
            def _thread_lock(self):
                """Get the lock specific to this collection.

                Since locks support the context manager protocol, this method
                can typically be invoked directly as part of a ``with`` statement.
                """
                return type(self)._locks[self._lock_id]

            cls._thread_lock = _thread_lock
            cls._threading_support_is_active = True
        else:
            raise ValueError("This class does not support multithreaded execution.")

    @classmethod
    def disable_multithreading(cls):
        """Disable all safety checks and thread locks required for thread safety.

        The mutex locks required to enable multithreading introduce nontrivial performance
        costs, so they can be disabled for classes that support it.

        """
        cls._thread_lock = _NullContext()
        cls._threading_support_is_active = False

    @property
    def _lock_id(self):
        raise NotImplementedError(
            "Backends must implement the _lock_id property to support multithreaded "
            "execution. This property should return a hashable unique identifier for "
            "all collections that will be used to maintain a resource-specific "
            "set of locks."
        )

    @property
    @abstractmethod
    def _backend(self):
        """str: The backend associated with a given collection.

        This property is abstract to enforce that subclasses implement it.
        Since it's only internal, subclasses can safely override it with just a
        raw attribute; this property just serves as a way to enforce the
        abstract API for subclasses.
        """
        pass

    @classmethod
    def _from_base(cls, data, **kwargs):
        r"""Dynamically resolve the type of object to the corresponding synced collection.

        This method assumes that ``data`` has already been validated. This assumption
        can always be met, since this method should only be called internally by
        other methods that modify the internal collection data. While this requirement
        does require that all calling methods be responsible for validation, it
        confers significant performance benefits because it can instruct any invoked
        class constructors not to validate, which is especially important for nested
        collections.

        Parameters
        ----------
        data : Collection
            Data to be converted from base type.
        \*\*kwargs
            Any keyword arguments to pass to the collection constructor.

        Returns
        -------
        Collection
            Synced object of corresponding base type.

        Notes
        -----
        This method relies on the internal registry of subclasses populated by
        :meth:`~.__init_subclass__` and the :meth:`is_base_type` method to
        determine the subclass with the appropriate backend and data type. Once
        an appropriate type is determined, that class's constructor is called.
        Since this method relies on the constructor and other methods, it can
        be concretely implemented here rather than requiring subclass
        implementations.

        """
        if _collection_resolver.get_type(data) == "COLLECTION":
            for base_cls in SyncedCollection.registry[cls._backend]:
                if base_cls.is_base_type(data):
                    return base_cls(data=data, _validate=False, **kwargs)
        return _convert_numpy(data)

    @abstractmethod
    def _to_base(self):
        """Dynamically resolve the synced collection to the corresponding base type.

        This method should not load the data from the underlying resource, it
        should simply converts the current in-memory representation of a synced
        collection to its naturally corresponding unsynced collection type.

        Returns
        -------
        Collection
            An equivalent unsynced collection satisfying :meth:`is_base_type`.

        """
        pass

    @classmethod
    @abstractmethod
    def is_base_type(cls, data):
        """Check whether data is of the same base type (such as list or dict) as this class.

        Parameters
        ----------
        data : Any
            The input data to test.

        Returns
        -------
        bool
            Whether or not the object can be converted into this synced collection type.

        """
        pass

    @abstractmethod
    def _load_from_resource(self):
        """Load data from underlying backend.

        This method must be implemented for each backend. Backends may choose
        to return ``None``, signaling that no modification should be performed
        on the data in memory. This mode is useful for backends where the underlying
        resource (e.g. a file) may not initially exist, but can be transparently
        created on save.

        Returns
        -------
        Collection or None
            An equivalent unsynced collection satisfying :meth:`is_base_type` that
            contains the data in the underlying resource (e.g. a file).

        """
        pass

    @abstractmethod
    def _save_to_resource(self):
        """Save data to the backend.

        This method must be implemented for each backend.
        """
        pass

    def _save(self):
        """Save the data to the backend.

        This method encodes the recursive logic required to handle the saving of
        nested collections. For a collection contained within another collection,
        only the parent is ever responsible for storing the data. This method
        handles the appropriate recursive calls, then farms out the actual writing
        to the abstract method :meth:`~._save_to_resource`.
        """
        if not self._suspend_sync:
            if self._root is None:
                self._save_to_resource()
            else:
                self._root._save()

    @abstractmethod
    def _update(self, data):
        """Update the in-memory representation to match the provided data.

        The purpose of this method is to update the SyncedCollection to match
        the data in the underlying resource.  The result of calling this method
        should be that ``self == data``. The reason that this method is
        necessary is that SyncedCollections can be nested, and nested
        collections must also be instances of SyncedCollection so that
        synchronization occurs even when nested structures are modified.
        Recreating the full nested structure every time data is reloaded from
        file is highly inefficient, so this method performs an in-place update
        that only changes entries that need to be changed.

        Parameters
        ----------
        data : Collection
            A collection satisfying :meth:`is_base_type`.

        """
        pass

    def _load(self):
        """Load the data from the backend.

        This method encodes the recursive logic required to handle the loading of
        nested collections. For a collection contained within another collection,
        only the root is ever responsible for loading the data. This method
        handles the appropriate recursive calls, then farms out the actual reading
        to the abstract method :meth:`~._load_from_resource`.
        """
        if not self._suspend_sync:
            if self._root is None:
                data = self._load_from_resource()
                with self._suspend_sync:
                    self._update(data)
            else:
                self._root._load()

    def _validate(self, data):
        """Validate the input data.

        Parameters
        ----------
        data : Collection
            An collection satisfying :meth:`is_base_type`.

        """
        for validator in self._all_validators:
            validator(data)

    # The following methods share a common implementation for
    # all data structures and regardless of backend.

    def __getitem__(self, key):
        self._load()
        return self._data[key]

    def __delitem__(self, key):
        with self._load_and_save:
            del self._data[key]

    def __iter__(self):
        self._load()
        return iter(self._data)

    def __len__(self):
        self._load()
        return len(self._data)

    def __call__(self):
        """Get an equivalent but unsynced object of the base data type.

        Returns
        -------
        Collection
            An equivalent unsynced collection satisfying :meth:`is_base_type`.

        """
        self._load()
        return self._to_base()

    def __eq__(self, other):
        self._load()
        if isinstance(other, type(self)):
            return self() == other()
        else:
            return self() == other

    def __repr__(self):
        self._load()
        return repr(self._data)

    def __str__(self):
        self._load()
        return str(self._data)



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Defines a buffering protocol for :class:`~.SyncedCollection` objects.

Depending on the choice of backend, synchronization may be an expensive process.
In that case, it can be helpful to allow many in-memory modifications to occur
before any synchronization is attempted. Since many collections could be pointing
to the same underlying resource, maintaining proper data coherency across different
instances requires careful consideration of how the data is stored. The appropriate
buffering methods can differ for different backends; as a result, the basic
interface simply lays out the API for buffering and leaves implementation
details for specific backends to handle. Judicious use of buffering can
dramatically speed up code paths that might otherwise involve, for instance,
heavy I/O. The specific buffering mechanism must be implemented by each backend
since it depends on the nature of the underlying data format.

All buffered collections expose a local context manager for buffering. In
addition, each backend exposes a context manager
:meth:`BufferedCollection.buffer_backend` that indicates to all buffered
collections of that backend that they should enter buffered mode. These context
managers may be nested freely, and buffer flushes will occur when all such
managers have been exited.

.. code-block:: python

    with collection1.buffered:
        with type(collection1).buffer_backend:
            collection2['foo'] = 1
            collection1['bar'] = 1
            # collection2 will flush when this context exits.

        # This operation will write straight to the backend.
        collection2['bar'] = 2

        # collection1 will flush when this context exits.
"""

import logging
from inspect import isabstract

from .. import SyncedCollection
from ..utils import _CounterFuncContext

logger = logging.getLogger(__name__)


class BufferedCollection(SyncedCollection):
    """A :class:`~.SyncedCollection` defining an interface for buffering.

    **The default behavior of this class is not to buffer.** This class simply
    defines an appropriate interface for buffering behavior so that client code
    can rely on these methods existing, e.g. to be able to do things like ``with
    collection.buffered: ...``. This feature allows client code to indicate to the
    collection when it is safe to buffer reads and writes, which usually means
    guaranteeing that the synchronization destination (e.g. an underlying file
    or database entry) will not be modified by other processes concurrently
    with the set of operations within the buffered block. However, in the
    default case the result of this will be a no-op and all data will be
    immediately synchronized with the backend.

    The BufferedCollection overrides the :meth:`~._load` and
    :meth:`~._save` methods to check whether buffering is enabled
    or not. If not, the behavior is identical to the parent class. When in buffered
    mode, however, the BufferedCollection introduces two additional hooks that
    can be overridden by subclasses to control how the collection behaves while buffered:

        - :meth:`~._load_from_buffer`: Loads data while in buffered mode and returns
          it in an object satisfying
          :meth:`~synced_collections.data_types.synced_collection.SyncedCollection.is_base_type`.
          The default behavior is to simply call
          :meth:`~._load_from_resource`.
        - :meth:`~._save_to_buffer`: Stores data while in buffered mode. The default behavior
          is to simply call
          :meth:`~._save_to_resource`.

    **Thread safety**

    Whether or not buffering is thread safe depends on the buffering method used. In
    general, both the buffering logic and the data type operations must be
    thread safe for the resulting collection type to be thread safe.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.buffered = _CounterFuncContext(self._flush)

    @classmethod
    def __init_subclass__(cls):
        """Register subclasses for the purpose of global buffering.

        Each subclass has its own means of buffering and must be flushed.
        """
        super().__init_subclass__()
        if not isabstract(cls):
            cls._buffer_context = _CounterFuncContext(cls._flush_buffer)

    @classmethod
    def buffer_backend(cls):
        """Enter context to buffer all operations for this backend."""
        return cls._buffer_context

    @classmethod
    def backend_is_buffered(cls):
        """Check if this backend is currently buffered."""
        return bool(cls._buffer_context)

    def _save(self):
        """Synchronize data with the backend but buffer if needed.

        This method is identical to the SyncedCollection implementation for
        `sync` except that it determines whether data is actually synchronized
        or instead written to a temporary buffer based on the buffering mode.
        """
        if not self._suspend_sync:
            if self._root is None:
                if self._is_buffered:
                    self._save_to_buffer()
                else:
                    self._save_to_resource()
            else:
                self._root._save()

    def _load(self):
        """Load data from the backend but buffer if needed.

        This method is identical to the :class:`~.SyncedCollection`
        implementation except that it determines whether data is actually
        synchronized or instead read from a temporary buffer based on the
        buffering mode.
        """
        if not self._suspend_sync:
            if self._root is None:
                if self._is_buffered:
                    data = self._load_from_buffer()
                else:
                    data = self._load_from_resource()
                with self._suspend_sync:
                    self._update(data)
            else:
                self._root._load()

    def _save_to_buffer(self):
        """Store data in buffer.

        By default, this method simply calls :meth:`~._save_to_resource`. Subclasses
        must implement specific buffering strategies.
        """
        self._save_to_resource()

    def _load_from_buffer(self):
        """Store data in buffer.

        By default, this method simply calls :meth:`~._load_from_resource`. Subclasses
        must implement specific buffering strategies.

        Returns
        -------
        Collection
            An equivalent unsynced collection satisfying
            :meth:`~synced_collections.data_types.synced_collection.SyncedCollection.is_base_type` that
            contains the buffered data. By default, the buffered data is just the
            data in the resource.

        """  # noqa: E501
        self._load_from_resource()

    @property
    def _is_buffered(self):
        """Check if we should write to the buffer or not."""
        return self.buffered or type(self)._buffer_context

    def _flush(self):
        """Flush data associated with this instance from the buffer."""
        pass

    @classmethod
    def _flush_buffer(self):
        """Flush all data in this class's buffer."""
        pass



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""A standardized buffering implementation for file-based backends.

The buffering method implemented here involves a single buffer of references to
in-memory objects containing data. These objects are the base types of a given
:class:`~.SyncedCollection` type, e.g. a dict for all dict-like collections,
and are the underlying data stores for those types. This buffering method
exploits the fact that all mutable collection types in Python are references,
so modifying one such collection results in modifying all of them, thereby
removing any need for more complicated synchronization protocols.
"""

from ..errors import MetadataError
from .file_buffered_collection import FileBufferedCollection


class SharedMemoryFileBufferedCollection(FileBufferedCollection):
    """A :class:`~.SyncedCollection` that defers all I/O when buffered.

    This class extends the :class:`~.FileBufferedCollection` and implements a
    concrete storage mechanism in which collections store a reference to their
    data in a buffer. This method takes advantage of the reference-based semantics
    of built-in Python mutable data types like dicts and lists. All collections
    referencing the same file are pointed to the same underlying data store in
    buffered mode, allowing all changes in one to be transparently reflected in
    the others. To further improve performance, the buffer size is determined
    only based on the number of modified collections stored, not the total number.
    As a result, the maximum capacity is only reached when a large number of
    modified collections are stored, and unmodified collections are only removed
    from the buffer when a buffered context is exited (rather than when buffer
    capacity is exhausted). See the Warnings section for more information.

    The buffer size and capacity for this class is measured in the total number
    of collections stored in the buffer that have undergone any modifications
    since their initial load from disk. A sequence of read-only operations will
    load data into the buffer, but the apparent buffer size will be zero.

    .. note::
        Important note for subclasses: This class should be inherited before
        any other collections. This requirement is due to the extensive use of
        multiple inheritance: since this class is designed to be combined with
        other :class:`~.SyncedCollection` types without making those types aware
        of buffering behavior, it transparently hooks into the initialization
        process, but this is dependent on its constructor being called before
        those of other classes.

    **Thread safety**

    This buffering method is thread safe. This thread safety is independent of the
    safety of an individual collection backend; the backend must support thread
    safe writes to the underlying resource in order for a buffered version using
    this class to be thread safe for general use. The thread safety guaranteed
    by this class only concerns buffer reads, writes, and flushes. All these
    operations are serialized because there is no way to prevent one collection
    from triggering a flush while another still thinks its data is in the cache;
    however, this shouldn't be terribly performance-limiting since in buffered
    mode we're avoiding I/O anyway and that's the only thing that can be effectively
    parallelized here.

    Parameters
    ----------
    filename : str, optional
        The filename of the associated JSON file on disk (Default value = None).

    Warnings
    --------
    - Although it can be done safely, in general modifying two different collections
      pointing to the same underlying resource while both are in different buffering
      modes is unsupported and can lead to undefined behavior. This class makes a
      best effort at performing safe modifications, but it is possible to construct
      nested buffered contexts for different objects that can lead to an invalid
      buffer state, or even situations where there is no obvious indicator of what
      is the canonical source of truth. In general, if you need multiple objects
      pointing to the same resource, it is **strongly** recommeneded to work with
      both of them in identical buffering states at all times.
    - This buffering method has no upper bound on the buffer size if all
      operations on buffered objects are read-only operations. If a strict upper bound
      is required, for instance due to strict virtual memory limits on a given system,
      use of the :class:`~.SerializedFileBufferedCollection` will allow limiting
      the total memory usage of the process.

    """

    _BUFFER_CAPACITY = 1000  # The number of collections to store in the buffer.

    def _flush(self, force=False):
        """Save buffered changes to the underlying file.

        Parameters
        ----------
        force : bool
            If True, force a flush even in buffered mode (defaults to False). This
            parameter is used when the buffer is filled to capacity.

        Raises
        ------
        MetadataError
            If any file is detected to have changed on disk since it was
            originally loaded into the buffer and modified.

        """
        # Different files in the buffer can be safely flushed simultaneously,
        # but a given file can only be flushed on one thread at once.
        if not self._is_buffered or force:
            try:
                cached_data = type(self)._buffer[self._filename]
            except KeyError:
                # If we got to this point, it means that another collection
                # pointing to the same underlying resource flushed the buffer.
                # If so, then the data in this instance is still pointing to
                # that object's data store. If this was a force flush, then
                # the data store is still the cached data, so we're fine. If
                # this wasn't a force flush, then we have to reload this
                # object's data so that it will stop sharing data with the
                # other instance.
                if not force:
                    self._data.clear()
                    self._update(self._load_from_resource())
            else:
                # If the contents have not been changed since the initial read,
                # we don't need to rewrite it.
                try:
                    # Validate that the file hasn't been changed by
                    # something else.
                    if cached_data["modified"]:
                        if cached_data["metadata"] != self._get_file_metadata():
                            raise MetadataError(self._filename, cached_data["contents"])
                        self._save_to_resource()
                finally:
                    # Whether or not an error was raised, the cache must be
                    # cleared to ensure a valid final buffer state, unless
                    # we're force flushing in which case we never delete, but
                    # take note that the data is no longer modified relative to
                    # its representation on disk.
                    if cached_data["modified"]:
                        type(self)._CURRENT_BUFFER_SIZE -= 1
                    if not force:
                        del type(self)._buffer[self._filename]
                    else:
                        # Have to update the metadata on a force flush because
                        # we could modify this item again later, leading to
                        # another (possibly forced) flush afterwards that will
                        # appear invalid if the metadata isn't updated to the
                        # metadata after the current flush.
                        cached_data["metadata"] = self._get_file_metadata()
                        cached_data["modified"] = False
        else:
            # If this object is still buffered _and_ this wasn't a force flush,
            # that implies a nesting of buffered contexts in which another
            # collection pointing to the same data flushed the buffer. This
            # object's data will still be pointing to that one, though, so the
            # safest choice is to reinitialize its data from scratch.
            with self._suspend_sync:
                self._data = {
                    key: self._from_base(data=value, parent=self)
                    for key, value in self._to_base().items()
                }

    def _load(self):
        """Load data from the backend but buffer if needed.

        Override the base buffered method to skip the _update and to let
        _load_from_buffer happen "in place."
        """
        if not self._suspend_sync:
            if self._root is None:
                if self._is_buffered:
                    self._load_from_buffer()
                else:
                    data = self._load_from_resource()
                    with self._suspend_sync:
                        self._update(data)
            else:
                self._root._load()

    def _save_to_buffer(self):
        """Store data in buffer.

        See :meth:`~._initialize_data_in_buffer` for details on the data stored
        in the buffer and the integrity checks performed.
        """
        type(self)._buffered_collections[id(self)] = self

        # Since one object could write to the buffer and trigger a flush while
        # another object was found in the buffer and attempts to proceed
        # normally, we have to serialize this whole block. In theory we might
        # be safe without it because the only operations that should reach this
        # point without already being locked are destructive operations (clear,
        # reset) that don't use the :meth:`_load_and_save` context, and for
        # those the writes will be automatically serialized because Python
        # dicts are thread-safe because of the GIL. However, it's best not to
        # depend on the thread-safety of built-in containers.
        with self._buffer_lock:
            if self._filename in type(self)._buffer:
                # Always track all instances pointing to the same data.

                # If all we had to do is set the flag, it could be done without any
                # check, but we also need to increment the number of modified
                # items, so we may as well do the update conditionally as well.
                if not type(self)._buffer[self._filename]["modified"]:
                    type(self)._buffer[self._filename]["modified"] = True
                    type(self)._CURRENT_BUFFER_SIZE += 1
            else:
                self._initialize_data_in_buffer(modified=True)
                type(self)._CURRENT_BUFFER_SIZE += 1

            if type(self)._CURRENT_BUFFER_SIZE > type(self)._BUFFER_CAPACITY:
                type(self)._flush_buffer(force=True)

    def _load_from_buffer(self):
        """Read data from buffer.

        See :meth:`~._initialize_data_in_buffer` for details on the data stored
        in the buffer and the integrity checks performed.

        Returns
        -------
        Collection
            A collection of the same base type as the SyncedCollection this
            method is called for, corresponding to data loaded from the
            underlying file.

        """
        super()._load_from_buffer()

        # Set local data to the version in the buffer.
        self._data = type(self)._buffer[self._filename]["contents"]

    def _initialize_data_in_buffer(self, modified=False):
        """Create the initial entry for the data in the cache.

        Stores the following information:
            - The metadata provided by :meth:`~._get_file_metadata`. Used to
              check if a file has been modified on disk since it was loaded
              into the buffer.
            - A flag indicating whether any operation that saves to the buffer
              has occurred, e.g. a ``__setitem__`` call. This flag is used to
              determine what collections need to be saved to disk when
              flushing.

        Parameters
        ----------
        modified : bool, optional
            Whether or not the data has been modified from the version on disk
            (Default value = False).

        """
        metadata = self._get_file_metadata()
        type(self)._buffer[self._filename] = {
            "contents": self._data,
            "metadata": metadata,
            "modified": modified,
        }

    @classmethod
    def _flush_buffer(cls, force=False):
        """Flush the data in the file buffer.

        Parameters
        ----------
        force : bool
            If True, force a flush even in buffered mode (defaults to False). This
            parameter is used when the buffer is filled to capacity.

        Returns
        -------
        dict
            Mapping from filenames to errors that occured while flushing data.

        Raises
        ------
        BufferedError
            If there are any issues with flushing the data.

        """
        return super()._flush_buffer(force=force, retain_in_force=True)
